// -*- Mode: C++; tab-width: 2; -*-
// vi: set ts=2:
//
// $Id: logStream.C,v 1.39.4.1 2005/06/10 20:47:43 oliver Exp $
//

#include <limits>
#include <string>
#include <BALL/COMMON/logStream.h>
#include <BALL/CONCEPT/notification.h>

#define BUFFER_LENGTH 32768

#ifdef BALL_HAS_ANSI_IOSTREAM
#	define BALL_IOS std::basic_ios<char>
#	define BALL_OSTREAM std::ostream
#else
#	define BALL_IOS std::ios
#	define BALL_OSTREAM std::ostream
#endif

namespace BALL 
{
	// at this point, it is not yet possible to
	// include BALL/COMMON/limits.h (which were a 
	// much nicer solution...). Ugly header dependencies...
	const int LogStreamBuf::MIN_LEVEL = INT_MIN;
	const int LogStreamBuf::MAX_LEVEL = INT_MAX;
	const Time LogStreamBuf::MAX_TIME = INT_MAX;

	LogStreamBuf::LogStreamBuf() 
		: std::streambuf(),
			pbuf_(0),
			loglines_(),
			level_(0),
			tmp_level_(0),
			stream_list_(),
			incomplete_line_()
	{
		pbuf_ = new char [BUFFER_LENGTH];
		std::streambuf::setp(pbuf_, pbuf_ + BUFFER_LENGTH - 1);
	}
		
	LogStreamBuf::~LogStreamBuf() 
	{
		sync();

		delete [] pbuf_;
	}

	void LogStreamBuf::dump(std::ostream& stream) 
	{
		char buf[BUFFER_LENGTH];
		Size line;
		for (line = (Size)loglines_.size(); line > 0; --line) 
		{
			strftime(&(buf[0]), BUFFER_LENGTH - 1, "%d.%m.%Y %H:%M:%S ", localtime(&(loglines_[line - 1].time)));
			stream << buf << "[" << loglines_[line - 1].level
						 << "]:" << loglines_[line - 1].text.c_str() << std::endl;
		}
	}
 
	int LogStreamBuf::sync() 
	{
		static char buf[BUFFER_LENGTH];

		// sync our streambuffer...
		if (pptr() != pbase()) 
		{
				
			char*	line_start = pbase();
			char*	line_end = pbase();
			
			while (line_end <= pptr())
			{
				// search for the first end of line
				for (; line_end < pptr() && *line_end != '\n'; line_end++);

				if (line_end >= pptr()) 
				{
					// Copy the incomplete line to the incomplete_line_ buffer
					size_t length = line_end - line_start + 1;
					length = std::max(length, (size_t)(BUFFER_LENGTH - 1));
					strncpy(&(buf[0]), line_start, length);
					buf[line_end - line_start] = '\0';
					incomplete_line_ += &(buf[0]);

					// mark everything as read
					line_end = pptr() + 1;
				} 
				else 
				{
					memcpy(&(buf[0]), line_start, line_end - line_start + 1);
					buf[line_end - line_start] = '\0';
						
					// assemble the string to be written
					// (consider leftovers of the last buffer from incomplete_line_)
					string outstring = incomplete_line_;
					incomplete_line_ = "";
					outstring += &(buf[0]);

					// if there are any streams in our list, we
					// copy the line into that streams, too and flush them
					std::list<StreamStruct>::iterator list_it = stream_list_.begin();
					for (; list_it != stream_list_.end(); ++list_it)
					{
						// if the stream is open for that level, write to it...
						if ((list_it->min_level <= tmp_level_) && (list_it->max_level >= tmp_level_))
						{
							*(list_it->stream) << expandPrefix_(list_it->prefix, tmp_level_, time(0)).c_str()
																 << outstring.c_str() << std::endl;
							if (list_it->target != 0)
							{
								list_it->target->notify();
							}
						}
					}
			
					// update the line pointers (increment both)
					line_start = ++line_end;
					
					// remove cr/lf from the end of the line				
					while (outstring[outstring.size() - 1] == 10 || outstring[outstring.size() - 1] == 13)
					{
						std::string::iterator p = outstring.end();
						p--;
						outstring.erase(p);
					}
		
					// store the line 
					Logline	logline;

					logline.text = outstring;
					logline.level = tmp_level_;
					logline.time = time(0);
			
					// store the new line
					loglines_.push_back(logline);

					// reset tmp_level_ to the previous level
					// (needed for LogStream::level() only)
					tmp_level_ = level_;
				}
			}

			// remove all processed lines from the buffer
			pbump((int)(pbase() - pptr()));
		}
	
		return 0;
	}

	string LogStreamBuf::expandPrefix_
		(const string& prefix, int level, Time time) const
	{
		string::size_type	index = 0;
		Size copied_index = 0;
		string result("");

		while ((index = prefix.find("%", index)) != string::npos) 
		{
			// append any constant parts of the string to the result
			if (copied_index < index) 
			{
				result.append(prefix.substr(copied_index, index - copied_index));
				copied_index = (Index)index;
			}
			
			if (index < prefix.size()) 
			{
				char	buffer[64];
				char*	buf = &(buffer[0]);

				switch (prefix[index + 1]) 
				{
					case '%': // append a '%' (escape sequence)
						result.append("%");
						break;

					case 'l': // append the loglevel
						sprintf(buf, "%d", level);
						result.append(buf);
						break;

					case 'y':	// append the message type (error/warning/information)
						if (level >= LogStream::ERROR_LEVEL) 
						{
							result.append("ERROR");
						}
						else 
						{
							if (level >= LogStream::WARNING_LEVEL) 
							{
								result.append("WARNING");
							}
							else 
							{
								if (level >= LogStream::INFORMATION_LEVEL) 
								{
									result.append("INFORMATION");
								}
								else 
								{
									result.append("LOG");
								}
							}
						}
						break;

					case 'T':	// time: HH:MM:SS
						strftime(buf, BUFFER_LENGTH - 1, "%H:%M:%S", localtime(&time));
						result.append(buf);
						break;

					case 't': // time: HH:MM	
						strftime(buf, BUFFER_LENGTH - 1, "%H:%M", localtime(&time));
						result.append(buf);
						break;

					case 'D':	// date: DD.MM.YYYY
						strftime(buf, BUFFER_LENGTH - 1, "%d.%m.%Y", localtime(&time));
						result.append(buf);
						break;

					case 'd':	// date: DD.MM.
						strftime(buf, BUFFER_LENGTH - 1, "%d.%m.", localtime(&time));
						result.append(buf);
						break;

					case 'S':	// time+date: DD.MM.YYYY, HH:MM:SS
						strftime(buf, BUFFER_LENGTH - 1, "%d.%m.%Y, %H:%M:%S", localtime(&time));
						result.append(buf);
						break;

					case 's':	// time+date: DD.MM., HH:MM
						strftime(buf, BUFFER_LENGTH - 1, "%d.%m., %H:%M", localtime(&time));
						result.append(buf);
						break;

					default:
						break;
				}
				index += 2;
				copied_index += 2;
			}
		}

		if (copied_index < prefix.size()) 
		{
			result.append(prefix.substr(copied_index, prefix.size() - copied_index));
		}

		return result;
	}

	LogStreamNotifier::LogStreamNotifier()
	{
	}

	LogStreamNotifier::LogStreamNotifier(const LogStream::Target& target)
	{
		NotificationRegister(*this, const_cast<LogStream::Target &>(target));
	}

	LogStreamNotifier::~LogStreamNotifier()
	{
		NotificationUnregister(*this);
	}

	void LogStreamNotifier::notify() const
	{
		Notify(const_cast<LogStreamNotifier&>(*this));
	}

	// keep the given buffer	
	LogStream::LogStream(LogStreamBuf* buf, bool delete_buf, bool associate_stdio)
		: BALL_IOS(buf),
			BALL_OSTREAM(buf),
			delete_buffer_(delete_buf),
			disable_output_(false)
	{
		if (associate_stdio == true) 
		{
			// associate cout to informations and warnings,
			// cerr to errors by default
			insert(std::cout, INFORMATION_LEVEL, ERROR_LEVEL - 1);
			insert(std::cerr, ERROR_LEVEL);
		}
	}

	// remove the streambuffer
	LogStream::~LogStream()
	{
		if (delete_buffer_)
		{
			delete rdbuf();
		}
	}
	
	void LogStream::clear()
	{
		rdbuf()->loglines_.clear();
	}

	void LogStream::insert(std::ostream& stream, int min_level, int max_level) 
	{
		// return if no LogStreamBuf is defined!
		if (rdbuf() == 0)
		{
			return;
		}
			
		// first, check whether the stream is already associated (avoid
		// multiple insertions)																																	
		std::list<LogStreamBuf::StreamStruct>::iterator list_it = rdbuf()->stream_list_.begin();
		for (; list_it != rdbuf()->stream_list_.end(); ++list_it) 
		{
			if ((*list_it).stream == &stream) 
			{
				return;
			}
		}

		// we didn`t find it - create a new entry in the list
		LogStreamBuf::StreamStruct s_struct;
		s_struct.min_level = min_level;
		s_struct.max_level = max_level;
		s_struct.stream = &stream;
		rdbuf()->stream_list_.push_back(s_struct);
	}

	void LogStream::remove(std::ostream& stream) 
	{
		// return if no LogStreamBuf is defined!
		if (rdbuf() == 0)
		{
			return;
		}
			
		// find the stream in the LogStreamBuf's list
		std::list<LogStreamBuf::StreamStruct>::iterator	list_it 
			= rdbuf()->stream_list_.begin();
		for (; list_it != rdbuf()->stream_list_.end(); ++list_it)
		{
			if (list_it->stream == &stream) 
			{
				// remove the stream - iterator becomes invalid, so exit
				rdbuf()->stream_list_.erase(list_it);
				break;
			}
		}
		
		// if the stream is not found nothing happens!		
	}

	void LogStream::insertNotification(const std::ostream& s, const LogStream::Target& target)
	{
		// return if no LogStreamBuf is defined!
		if (rdbuf() == 0)
		{
			return;
		}
			
		// find the stream in the LogStreamBuf's list
		std::list<LogStreamBuf::StreamStruct>::iterator	list_it = rdbuf()->stream_list_.begin();
		for (; list_it != rdbuf()->stream_list_.end(); ++list_it)
		{
			if (list_it->stream == &s) 
			{
				// set the notification target
				list_it->target = new LogStreamNotifier(const_cast<LogStream::Target&>(target));
				break;
			}
		}
		
		// if the stream is not found nothing happens!		
	}

	
	void LogStream::removeNotification(const std::ostream& s) 
	{
		// return if no LogStreamBuf is defined!
		if (rdbuf() == 0)
		{
			return;
		}
			
		// find the stream in the LogStreamBuf's list
		std::list<LogStreamBuf::StreamStruct>::iterator	list_it = rdbuf()->stream_list_.begin();
		for (; list_it != rdbuf()->stream_list_.end(); ++list_it)
		{
			if (list_it->stream == &s) 
			{
				// destroy and remove the notification target
				delete list_it->target;
				list_it->target = 0;
				break;
			}
		}
		
		// if the stream is not found nothing happens!		
	}

	void LogStream::setMinLevel(const std::ostream& stream, int level) 
	{
		// return if no LogStreamBuf is defined!
		if (rdbuf() == 0)
		{
			return;
		}
			
		// find the stream in the LogStreamBuf's list
		std::list<LogStreamBuf::StreamStruct>::iterator	list_it = rdbuf()->stream_list_.begin();
		for (; list_it != rdbuf()->stream_list_.end(); ++list_it)
		{
			if ((*list_it).stream == &stream) 
			{
				// change the streams min_level and exit the loop
				(*list_it).min_level = level;
				break;
			}	
		}
	}

	void LogStream::setMaxLevel(const std::ostream& stream, int level) 
	{
		// return if no LogStreamBuf is defined!
		if (rdbuf() == 0)
		{
			return;
		}
			
		// find the stream in the LogStreamBuf's list:
		// iterate over the list until you find the stream`s pointer

		std::list<LogStreamBuf::StreamStruct>::iterator	list_it = rdbuf()->stream_list_.begin();
		for (; list_it != rdbuf()->stream_list_.end(); ++list_it)
		{
			if ((*list_it).stream == &stream) 
			{
				// change the streams max_level and exit the loop
				(*list_it).max_level = level;
				break;
			}
		}
	}
	
	void LogStream::setPrefix(const std::ostream& s, const string& prefix) 
	{
		// return if no LogStreamBuf is defined!
		if (rdbuf() == 0)
		{
			return;
		}
			
		// find the stream in the LogStreamBuf's list:
		// iterate over the list until you find the stream`s pointer
		std::list<LogStreamBuf::StreamStruct>::iterator	list_it = rdbuf()->stream_list_.begin();
		for (; list_it != rdbuf()->stream_list_.end(); ++list_it)
		{
			if ((*list_it).stream == &s) 
			{
				// change the streams max_level and exit the loop
				list_it->prefix = prefix;
				return;
			} 
		}
	}
	
	Size LogStream::getNumberOfLines(int min_level, int max_level) const  
	{
		// cast this to const, to access non const method rdbuf() which
		// is usually const, but declared non const
		LogStream*	non_const_this = const_cast<LogStream*>(this);

		// if rdbuf() is NULL, return
		if (non_const_this->rdbuf() == 0)
		{
			return 0;
		}

		// iterate over all loglines and count the lines of interest

		vector<LogStreamBuf::Logline>::iterator	it = non_const_this->rdbuf()->loglines_.begin();
		Size	count = 0;

		for (; it != non_const_this->rdbuf()->loglines_.end(); ++it) 
		{
			if ((*it).level >= min_level && (*it).level <= max_level)
			{
				count++;
			}
		}

		return count;
	}

	string LogStream::getLineText(const Index& index) const
	{
		if ((signed)getNumberOfLines() < index)
		{
			return "";
		}

		LogStream* non_const_this = const_cast<LogStream*>(this);

		if (non_const_this->rdbuf() == 0)
		{
			return "";
		}

		return non_const_this->rdbuf()->loglines_[index].text;	
	}

	int LogStream::getLineLevel(const Index& index) const
	{
		if ((signed)getNumberOfLines() < index)
		{
			return -1;
		}

		LogStream*	non_const_this = const_cast<LogStream*>(this);

		if (non_const_this->rdbuf() == 0)
		{
			return -1;
		}

		return non_const_this->rdbuf()->loglines_[index].level;	
	}


	time_t LogStream::getLineTime(const Index& index) const
	{
		if ((signed)getNumberOfLines() < index)
		{
			return 0;
		}

		LogStream*	non_const_this = const_cast<LogStream*>(this);

		if (non_const_this->rdbuf() == 0)
		{
			return 0;
		}

		return non_const_this->rdbuf()->loglines_[index].time;	
	}

	std::list<int>	LogStream::filterLines
		(int min_level, int max_level,
		 Time earliest, Time latest, const string& s) const
	{
		std::list<int>	list_indices;
		Position pos = 0;
		LogStreamBuf* log = const_cast<LogStream*>(this)->rdbuf();

		while (pos < log->loglines_.size() && 
					 log->loglines_[pos].time < earliest)
		{
			pos++;
		}
		while (pos < log->loglines_.size() && 
					 log->loglines_[pos].time <= latest)
		{
			if (log->loglines_[pos].level >= min_level &&
					log->loglines_[pos].level <= max_level)
			{
				if (s.length() > 0)
				{
					if (log->loglines_[pos].text.find(s, 0) != string::npos)
					{
						list_indices.push_back((int)pos);
					}
				}
				else
				{
					list_indices.push_back((int)pos);
				}
			}
			pos++;
		}
		return list_indices;
	}

	void LogStream::disableOutput()
		throw()
	{
		disable_output_ = true;
	}

	void LogStream::enableOutput()
		throw()
	{
		disable_output_ = false;
		std::ostream::flush();
	}

	bool LogStream::outputEnabled() const
		throw()
	{
		return disable_output_;
	}

	void LogStream::flush()
		throw()
	{
		if (disable_output_) return;

		std::ostream::flush();
	}

	// global default logstream
	LogStream	Log(new LogStreamBuf, true, true);

#	ifdef	BALL_NO_INLINE_FUNCTIONS
#		include <BALL/COMMON/logStream.iC>
#	endif

} // namespace BALL
